/*******************************************************************************
 * IBM Confidential
 * OCO Source Materials
 * IBM HTML Commons library Version 0.01 (IBM-HTML-Commons-0001)
 * (c) Copyright IBM Corporation 2006.
 * All rights reserved.
 * 
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been 
 * deposited with the U.S. Copyright Office.
 * 
 * IBM Confidential
 *     2005/11/7 - updated by Mei Yang (IBM Corp.) for Quick Print uat4i00000015
 *     2005/12/7 - updated by Mei Yang (IBM Corp.) for uat4i00000210
 *     2006/3/7 - updated by Mei Yang (IBM Corp.) for uat4i00000507 1
 *******************************************************************************/

package com.ibm.watson.commons.html.wrapper;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import com.ibm.watson.commons.html.parseutils.ParseCorrect;
import com.ibm.watson.commons.html.parseutils.ParseError;
import com.ibm.watson.commons.html.parseutils.ParseResult;

/**
 * A robust HTML parser, implemented as a static <code>parse</code> method,
 * returning a list of parsing chunks.
 * 
 * @author Pablo A. Duboue (<a
 *         href="mailto:duboue@us.ibm.com">duboue@us.ibm.com</a>)
 */
public class RobustHTMLParser {
	 private static final String copyright = "Licensed Materials - Property of IBM\n"
         + "IEHS-311\n"
         + "(c) Copyright IBM Corporation 2006.\n"
         + "All rights reserved.\n"
         + "US Government Users Restricted Right - Use, duplication or\n"
         + "disclosure restricted by GSA ADP Schedule Contract with\n"
         + "IBM Corporation.";

	private RobustHTMLParser() {
	}

	public static class Chunk {
	}

	public static class Tag extends Chunk {
		protected String tag;

		public Tag(String tag) {
			this.tag = tag.toLowerCase(Locale.ENGLISH);
		}

		/**
		 * @return Returns the tag.
		 */
		public String getTag() {
			return tag;
		}
	}

	public static class OpenTag extends Tag {
		Map params;

		public OpenTag(String tag, Map params) {
			super(tag);
			// this.params = Collections.unmodifiableMap(params);
			this.params = params;
		}

		/**
		 * @return Returns the params.
		 */
		public Map getParams() {
			return params;
		}

		// uat4i00000507
		public void setParams(Map params) {
			this.params = params;
		}
		// end of uat4i00000507
		
		protected String getParamsAsString() {
			StringBuffer buffer = new StringBuffer();
			Set s = params.entrySet();
			for (Iterator iter = s.iterator(); iter.hasNext();) {
				Map.Entry entry = (Map.Entry) iter.next();
				buffer.append(" ").append(entry.getKey()).append("=\"").append(
						entry.getValue()).append("\"");
			}
			return new String(buffer);
		}

		public String toString() {

			return "<" + tag + getParamsAsString() + ">";
		}
	}

	// TODO
	public static class CompleteTag extends OpenTag {

		public CompleteTag(String tag, Map params) {
			super(tag, params);
		}

		public String toString() {
			return "<" + tag + getParamsAsString() + "/>";
		}
	}

	public static class CloseTag extends Tag {
		public CloseTag(String tag) {
			super(tag);
		}

		public String toString() {
			return "</" + tag + ">";
		}
	}

	// TODO temporarily treat commet as a specail tag
	public static class CommentTag extends CloseTag {
		public CommentTag(String tag) {
			super(tag);
		}
	}
	
	public static class Text extends Chunk {
		String text;

		public Text(String text) {
			this.text = text;
		}

		/**
		 * @return Returns the text.
		 */
		public String getText() {
			return text;
		}

		/**
		 * @param text
		 *            The text to set.
		 */
		public void setText(String text) {
			this.text = text;
		}

		public String toString() {
			return text;
		}
	}

	public static List parse(String s) {
		List result = new LinkedList();
		int c = 0;
		while (c < s.length()) {
			// System.err.println("currently at: "+c);
			int posOpenBracket = c;
			boolean first = true;
			do {
				posOpenBracket = s.indexOf('<', posOpenBracket
						+ (first ? 0 : 1));
				first = false;
			} while (!(posOpenBracket < 0 || posOpenBracket == s.length() - 1) // no
					// more
					// tags
					&& Character.isWhitespace(s.charAt(posOpenBracket + 1)) // spurious
			// match
			);
			if (posOpenBracket < 0 || posOpenBracket == s.length() - 1) { // no
				// more
				// tags
				result.add(new Text(s.substring(c)));
				c = s.length();
			} else { // found a non-spurious bracket
				if (posOpenBracket != c) // commit text in between
					result.add(new Text(s.substring(c, posOpenBracket)));
				// parse the interior of the tag, till the closing bracket
				// extra attention to brackets inside attribute values
				ParseResult parsedTag = parseTag(s, posOpenBracket);

				if (parsedTag instanceof ParseError) {
					// add the segment as text and try to recover
					int recovery = parsedTag.getRemainderAsInt();
					if (recovery == -1)
						recovery = s.length();
					result.add(new Text(s.substring(posOpenBracket,
							recovery - 1)));
					c = recovery;
				} else {
					result.add(parsedTag.getParsedValue());
					c = parsedTag.getRemainderAsInt();
				}

			}
		}
		return result;
	}

	/**
	 * Parse an opening or closing tag.
	 * 
	 * @param s
	 *            the string being parsed
	 * @param bracketPos
	 *            the position where the open-bracket with a non-white character
	 *            following it was encountered
	 * @return the result of the parsing, an Tag enclosed in ParseCorrect, if
	 *         successfull, ParseError, otherwise.
	 */
	protected static ParseResult parseTag(String s, int bracketPos) {
		if (s.charAt(bracketPos + 1) == '/') { // close tag
			ParseResult newPos = consumeWhiteSpace(s, bracketPos + 2);
			ParseResult parsedName = parseName(s, newPos.getRemainderAsInt());
			if (parsedName instanceof ParseError)
				return new ParseError("Bad name", parsedName.getRemainder());
			ParseResult finalPos = consumeUntil(s, parsedName
					.getRemainderAsInt(),
					new Character[] { new Character('>') });
			if (parsedName instanceof ParseError)
				return new ParseError("Tag '" + parsedName.getParsedValue()
						+ "' without '>'", finalPos.getRemainder());
			return new ParseCorrect(new CloseTag((String) parsedName
					.getParsedValue()), new Integer(finalPos
					.getRemainderAsInt() + 1));
		} else if (s.charAt(bracketPos + 1) == '!') { // comment
			// has '--'?
			ParseResult newPos = consumeWhiteSpace(s, bracketPos + 2);
			int c = newPos.getRemainderAsInt();
			if (s.charAt(c) == '-' && s.charAt(c + 1) == '-') { // has dashes
				c += 2;
				ParseResult dashesPos = consumeUntil(s, c,
						new Character[] { new Character('-') });
				while ((dashesPos instanceof ParseCorrect)
						&& dashesPos.getRemainderAsInt() < s.length()
						&& s.charAt(dashesPos.getRemainderAsInt()) != '-') {
					c = dashesPos.getRemainderAsInt() + 1;
					dashesPos = consumeUntil(s, c,
							new Character[] { new Character('-') });
				}
				if (dashesPos instanceof ParseError
						|| dashesPos.getRemainderAsInt() == s.length())
					return new ParseError(
							"Unfinished comment run past end of file.",
							new Integer(s.length()));
				c = dashesPos.getRemainderAsInt() + 2; // skip double dashes
			}
			ParseResult finalPos = consumeUntil(s, c,
					new Character[] { new Character('>') });

//			return new ParseCorrect(new CloseTag("comment"), new Integer(
//					finalPos.getRemainderAsInt() + 1));
			return new ParseCorrect(new CommentTag("comment"), new Integer(
					finalPos.getRemainderAsInt() + 1));

		}
		// open tag
		ParseResult parsedName = parseName(s, bracketPos + 1);
		if (parsedName instanceof ParseError)
			return new ParseError("Bad name", parsedName.getRemainder());
		
		// uat4i00000210
		// TODO special process for script tag
		if ("script".equals(((String)parsedName.getParsedValue()).toLowerCase(Locale.ENGLISH))) {
			int endPos = s.indexOf("</", bracketPos);
			ParseResult finalPos = consumeUntil(s, endPos+8,
					new Character[] { new Character('>') });
			return new ParseCorrect(new CloseTag((String) parsedName
					.getParsedValue()), new Integer(finalPos.getRemainderAsInt() + 1));
		}
		// end of uat4i00000210
		
		Map headers = new HashMap();
		int c = parsedName.getRemainderAsInt();
		ParseResult parsedArg = parseArg(s, c);
		while (parsedArg instanceof ParseCorrect) {
			headers.putAll((Map) parsedArg.getParsedValue());
			c = parsedArg.getRemainderAsInt();
			parsedArg = parseArg(s, c);
		}
		ParseResult finalPos = consumeUntil(s, c,
				new Character[] { new Character('>') });
		return new ParseCorrect(new OpenTag((String) parsedName
				.getParsedValue(), headers), new Integer(finalPos
				.getRemainderAsInt() + 1));
	}

	/**
	 * Consume white space.
	 * 
	 * @param s
	 *            string being parsed
	 * @param pos
	 *            current position
	 * @return the new position
	 */
	protected static ParseResult consumeWhiteSpace(String s, int pos) {
		while (pos < s.length() && Character.isWhitespace(s.charAt(pos)))
			pos++;
		return new ParseCorrect(null, new Integer(pos));
	}

	/**
	 * Read a name (either tag or attribute). A name is any character different
	 * from whitespace, an equal sign, a &gt; and a /.
	 * 
	 * @param s
	 *            the string being parsed
	 * @param pos
	 *            the current position
	 * @return the name and the position past the name, error if the name is
	 *         empty.
	 */
	protected static ParseResult parseName(String s, int pos) {
		int start = pos;
		while (pos < s.length() && !Character.isWhitespace(s.charAt(pos))
				&& s.charAt(pos) != '=' && s.charAt(pos) != '>'
				&& s.charAt(pos) != '/')
			pos++;
		if (start == pos)
			return new ParseError("Empty name.", new Integer(pos + 1));
		return new ParseCorrect(s.substring(start, pos), new Integer(pos));
	}

	/**
	 * Consume until a delimiter
	 * 
	 * @param s
	 *            the string being parsed
	 * @param pos
	 *            current position
	 * @param delim
	 *            the list of character delimiters
	 * @return the position where one of the delimiters was found. Error
	 *         otherwise.
	 */
	protected static ParseResult consumeUntil(String s, int pos,
			Character[] delim) {
		Set delimSet = new HashSet(Arrays.asList(delim));
		while (pos < s.length()
				&& !delimSet.contains(new Character(s.charAt(pos))))
			pos++;
		if (pos == s.length())
			return new ParseError(null, new Integer(s.length()));
		return new ParseCorrect(null, new Integer(pos));
	}

	/**
	 * Parse an argument/value pair. It accepts empty attributes and attributes
	 * delimited with single, double and no quotes.
	 * 
	 * @param s
	 *            the string being parsed
	 * @param pos
	 *            current position
	 * @return A map containing the attribute value pair (the value is the empty
	 *         string for empty attributes).
	 */
	protected static ParseResult parseArg(String s, int pos) {
		ParseResult newPos = consumeWhiteSpace(s, pos);
		pos = newPos.getRemainderAsInt();
		ParseResult parsedName = parseName(s, newPos.getRemainderAsInt());
		if (parsedName instanceof ParseError)
			return new ParseError(parsedName.getRemainder());
		Map pair = new HashMap();
		Object key = parsedName.getParsedValue();
		if (key != null)
			key = ((String) key).toLowerCase(Locale.ENGLISH);
		pair.put(key, "");
		ParseResult equalPos = consumeWhiteSpace(s, parsedName
				.getRemainderAsInt());
		if (s.charAt(equalPos.getRemainderAsInt()) != '=')
			return new ParseCorrect(pair, equalPos.getRemainder());
		pos = equalPos.getRemainderAsInt() + 1;
		ParseResult valuePos = consumeWhiteSpace(s, pos);
		if (s.charAt(valuePos.getRemainderAsInt()) == '"'
				|| s.charAt(valuePos.getRemainderAsInt()) == '\'') {
			// has delim
			char delim = s.charAt(valuePos.getRemainderAsInt());
			ParseResult closeDelimPos = consumeUntil(s, valuePos
					.getRemainderAsInt() + 1, new Character[] { new Character(
					delim) });
			if (closeDelimPos instanceof ParseError)
				return new ParseError("Attribute value delimiter '" + delim
						+ "' not closed for attribute '"
						+ parsedName.getParsedValue() + "'", closeDelimPos
						.getRemainder());
			key = parsedName.getParsedValue();
			if (key != null)
				key = ((String) key).toLowerCase(Locale.ENGLISH);
			pair.put(key, s.substring(valuePos.getRemainderAsInt() + 1,
					closeDelimPos.getRemainderAsInt()));
			return new ParseCorrect(pair, new Integer(closeDelimPos
					.getRemainderAsInt() + 1));
		} else {
			// no delim, more complicated
			ParseResult endValuePos = consumeUntil(s, valuePos
					.getRemainderAsInt(), new Character[] { new Character('>'),
					new Character(' '), new Character('\t'),
					new Character('\n') });
			if (endValuePos instanceof ParseError)
				return new ParseError("Value not closed for attribute '"
						+ parsedName.getParsedValue() + "'", endValuePos
						.getRemainder());
			key = parsedName.getParsedValue();
			if (key != null)
				key = ((String) key).toLowerCase(Locale.ENGLISH);
			pair.put(key, s.substring(valuePos.getRemainderAsInt(), endValuePos
					.getRemainderAsInt()));
			return new ParseCorrect(pair, endValuePos.getRemainder());
		}
	}
}
