/*******************************************************************************
 * Copyright (c) 2007, 2008 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     2008/09/08 - updated by Hu Qiongkai(IBM Corp.) for Defect 644
 *     2008/09/12 - updated by Hu Qiongkai(IBM Corp.) for Defect 657 for quick print enhancement
 *     2008/09/22 - updated by Xu Yumei(IBM Corp.) for Filter print topics
 *     2008/10/28 - updated by Xu Yumei(IBM Corp.) for Single Print
 *     2008/10/31 - updated by Xu Yumei(IBM Corp.) for Filter print topics content
 *     2008/11/03 - updated by Xu Yumei(IBM Corp.) for quick print for anchor doc
 *******************************************************************************/
package org.eclipse.help.internal.webapp.data;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.eclipse.core.runtime.Preferences;
import org.eclipse.help.HelpSystem;
import org.eclipse.help.IToc;
import org.eclipse.help.ITopic;
import org.eclipse.help.IUAElement;
import org.eclipse.help.internal.HelpPlugin;
import org.eclipse.help.internal.base.HelpBasePlugin;
//import org.eclipse.help.internal.base.remote.RemoteHelp;
import org.eclipse.help.internal.base.BaseHelpSystem;
import org.eclipse.help.internal.search.HTMLDocParser;
import org.eclipse.help.internal.webapp.HelpWebappPlugin;
import org.eclipse.help.internal.xhtml.DynamicXHTMLProcessor;
//
import com.ibm.watson.commons.html.wrapper.RobustHTMLParser;
import com.ibm.watson.commons.html.wrapper.RobustHTMLParser.Chunk;
import com.ibm.watson.commons.html.wrapper.RobustHTMLParser.CommentTag;
import com.ibm.watson.commons.html.wrapper.RobustHTMLParser.OpenTag;
import com.ibm.watson.commons.html.wrapper.RobustHTMLParser.Tag;

/*
 * Used by the print jsp to access print-related data.
 */
public class PrintData extends RequestData {

	// where to inject the section numbers
	private static final Pattern PATTERN_HEADING = Pattern.compile("<body.*?>[\\s]*?([\\w])", Pattern.MULTILINE | Pattern.DOTALL | Pattern.CASE_INSENSITIVE); //$NON-NLS-1$

	// to normalize external links to new base href
	private static final Pattern PATTERN_LINK = Pattern.compile("(src|href)=\"(.*?\")", Pattern.MULTILINE | Pattern.DOTALL | Pattern.CASE_INSENSITIVE); //$NON-NLS-1$
	
	// Story 702: quick print enhancement
	private static Set images;

	private static final int defaultMaxConnections = 50;

	private static final int defaultMaxTopics = 100;

	private static int allowedMaxConnections;

	private static int allowedMaxTopics;
	
	// Number of topics generated so far
	private int topicsGenerated = 0;
	
	private int tocGenerated = 0;
	
	// Number of topics requested
	private int topicsRequested = 0;
	
	private int topicID = 1;
	
	private static final Object lock = new Object();
	
	private boolean isZOS = false;
	
	private int bufferSize = 0;
	
	String direction;
	
	private boolean confirmed;
	private boolean singlePrint = false;
	
	private static boolean initialized = false;

	/*
	 * Constructs the print data for the given request.
	 */
	public PrintData(ServletContext context, HttpServletRequest request, HttpServletResponse response) 
			throws ServletException, IOException {
		super(context, request, response);
		
		if (!isInitialized()) {
			initPreferences();
		}
		
		// init direction
		boolean isRTL = UrlUtil.isRTL(request, response);
		direction = isRTL ? "rtl" : "ltr";		
		
		String confirmString = request.getParameter("confirmed");
		if ((confirmString != null) && ("true".equals(confirmString))) {
			confirmed = true;
		} else {
			confirmed = false;
		}
		
		String isSinglePrint = request.getParameter("isSinglePrint");
		if ((isSinglePrint != null) && ("true".equals(isSinglePrint))) {
			singlePrint = true;
		}

		// check resource allocation
		synchronized (lock) {
			if (allowedMaxConnections <= 0) {
				RequestDispatcher rd = context.getRequestDispatcher("/advanced/printError.jsp"); //$NON-NLS-1$
				request.setAttribute("msg", "noConnection"); //$NON-NLS-1$ //$NON-NLS-2$
				rd.forward(request, response);
				allowedMaxConnections--;
				return;
			} else {
				allowedMaxConnections--;
			}
		}
		
		if (isZOS) {
			response.setBufferSize(bufferSize);
		}
		
		ITopic topic = getTopic();		// topic selected for print
		if (isSinglePrint()) {
			setTopicsRequested(1);
		}
		else {
			topicsRequested(topic);
		}
		
		if (getTopicsRequested() > allowedMaxTopics) {	
			if (!isConfirmed()) {
				// return resource
				synchronized (lock) {
					allowedMaxConnections++;
					lock.notifyAll();
				}				
				RequestDispatcher rd = context.getRequestDispatcher("/advanced/printConfirm.jsp?"+request.getQueryString());
				request.setAttribute("topicsRequested", String.valueOf(getTopicsRequested()));
				request.setAttribute("allowedMaxTopics", String.valueOf(allowedMaxTopics));
				rd.forward(request, response);
				return;
			}
			setTopicsRequested(allowedMaxTopics);  // reset topics requested to maxTopics
		}
	}

	public int getTopicsRequested() {
		return topicsRequested;
	}

	public void setTopicsRequested(int topicsRequested) {
		this.topicsRequested = topicsRequested;
	}
	
	public boolean isConfirmed() {
		return confirmed;
	}
	
	public boolean isSinglePrint() {
		return singlePrint;
	}

	public boolean isInitialized() {
		return initialized;
	}
	
	/*
	 * count topics num for quick print
	 */
	private void topicsRequested(ITopic topic) {
		if (topic.getHref() != null && topic.getHref().length() > 0) {
			topicsRequested++;
		}
		
		ITopic[] subtopics = EnabledTopicUtils.getEnabled(topic.getSubtopics());
		for (int i=0;i<subtopics.length;++i) {
			topicsRequested(subtopics[i]);
		}
	}
	
	/*
	 * init properties set in base preference.ini for quick print
	 */
	private void initPreferences() {		
		isZOS = preferences.isZOS();
		bufferSize = isZOS ? preferences.getResponseBufferSize() : 0;
		
		// set image types supported for print
		Preferences pref = HelpBasePlugin.getDefault().getPluginPreferences();
		String imageFormats = pref.getString("image"); //$NON-NLS-1$
		String[] formats = imageFormats.split("_"); //$NON-NLS-1$
		images = new HashSet(formats.length);
		for (int i = 0; i < formats.length; ++i) {
			images.add(formats[i]);
		}
		
		// set max connection numbers for concurrent access
		String maxConnections = pref.getString("maxConnections"); //$NON-NLS-1$
		if ((maxConnections == null) || ("".equals(maxConnections.trim()))) { //$NON-NLS-1$
			allowedMaxConnections = defaultMaxConnections;
		} else {
			try {
				allowedMaxConnections = Integer.parseInt(maxConnections);
			} catch (NumberFormatException e) {
				HelpWebappPlugin.logError(
						"Init maxConnections error. Set to default.", e); //$NON-NLS-1$
				allowedMaxConnections = defaultMaxConnections;
			}
		}
		
		// set max topics allowed to print in one request
		String maxTopics = pref.getString("maxTopics"); //$NON-NLS-1$
		if ((maxTopics == null) || ("".equals(maxTopics.trim()))) { //$NON-NLS-1$
			allowedMaxTopics = defaultMaxTopics;
		} else {
			try {
				allowedMaxTopics = Integer.parseInt(maxTopics);
			} catch (NumberFormatException e) {
				HelpWebappPlugin.logError(
						"Init maxTopics error. Set to default.", e); //$NON-NLS-1$
				allowedMaxTopics = defaultMaxTopics;
			}
		}
		
		initialized = true;
	}
	
	/*
	 * Returns the overall topic's title.
	 */
	public String getTitle() {
		return getTopic().getLabel();
	}

	/*
	 * Returns the href of the toc containing the topic(s) to print.
	 */
	public String getTocHref() {
		return getToc().getHref();
	}
	
	/*
	 * Returns the href of the root topic to print.
	 */
	public String getTopicHref() {
		return getTopic().getHref();
	}

	/*
	 * Generates and outputs a table of contents div with links.
	 */
	public void generateToc(Writer out) throws IOException, ServletException {
		try {
			out.write("<div id=\"toc\"  dir=\"" + direction + "\">"); //$NON-NLS-1$
			out.write("<h1>"); //$NON-NLS-1$
			out.write(getTitle());
			out.write("</h1>"); //$NON-NLS-1$
			out.write("<h2>"); //$NON-NLS-1$
			out.write(ServletResources.getString("TocHeading", request)); //$NON-NLS-1$
			out.write("</h2>"); //$NON-NLS-1$
			out.write("<div id=\"toc_content\" dir=\"" + direction + "\">"); //$NON-NLS-1$
			if (!isSinglePrint()) {
				ITopic topic = getTopic();
				
				String href = topic.getHref();
				if (href != null && href.length() > 0) {
					tocGenerated++;
				}

				ITopic[] subtopics = EnabledTopicUtils.getEnabled(topic.getSubtopics());
				topicID = 1;
				for (int i=0;i<subtopics.length;++i) {
					boolean generateSuc = generateToc(subtopics[i], String.valueOf(topicID), out);
					// added for filter
					if(generateSuc) 
						topicID++;
					// ends
				}
			}
			out.write("</div>"); //$NON-NLS-1$
			out.write("</div>"); //$NON-NLS-1$
		} catch (IOException e) {
			// return resource
			synchronized (lock) {
				allowedMaxConnections++;
				lock.notifyAll();
			}
			RequestDispatcher rd = context.getRequestDispatcher("/advanced/printError.jsp"); //$NON-NLS-1$
			request.setAttribute("msg", "ioException"); //$NON-NLS-1$ //$NON-NLS-2$
			rd.forward(request, response);
			return;
		}
		catch (Exception ex) {
			HelpWebappPlugin.logWarning("Print is terminated by an exception."); //$NON-NLS-1$
		}		
	}

	/*
	 * Auxiliary method for recursively generating table of contents div.
	 */
	private boolean generateToc(ITopic topic, String sectionId, Writer out) throws IOException {
		// filter the topic first
		if (!FilterData.filter(request.getSession(), topic)) {
			return false;
		}
		
		if(topic.getHref() == null || topic.getHref().length() == 0) {
			ITopic[] topics = getFilteredSubtopics(topic);
			if(topics.length == 0) {
				return false;
			}
		}
		//filter ends
		if (tocGenerated < allowedMaxTopics) {		// to ensure generating the right toc for printed topics
			out.write("<div dir=\"" + direction + "\" class=\"toc_" + (sectionId.length() > 2 ? "sub" : "") + "entry\">"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
			out.write(sectionId + ". " + "<a href=\"#section" + sectionId + "\">" + topic.getLabel() + "</a>"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
			
			String href = topic.getHref();
			if (href != null && href.length() > 0) {
				tocGenerated++;
			}
			
			ITopic[] subtopics = EnabledTopicUtils.getEnabled(topic.getSubtopics());
			for (int i=0;i<subtopics.length;++i) {
				String subsectionId = sectionId + "." + (i + 1); //$NON-NLS-1$
				generateToc(subtopics[i], subsectionId, out);
			}
			out.write("</div>"); //$NON-NLS-1$
			
			return true;
		}		
		return false;
	}
	
	//**********************Filter starts************************************
	/**
	 * Obtains children topics for a given navigation element. Topics from TOCs
	 * not matching enabled activities are filtered out.
	 * 
	 * @param navigationElement
	 * @return ITopic[]
	 */
	private ITopic[] getFilteredSubtopics(
			ITopic navigationElement) {
		List topics = getFilteredSubtopicList(navigationElement);
		return (ITopic[]) topics
				.toArray(new ITopic[topics.size()]);
	}
	/**
	 * Obtains children topics for a given navigation element. Topics from TOCs
	 * not matching enabled activities are filtered out.
	 * 
	 * @param navigationElement
	 * @return List of ITopicElement
	 */
	private List getFilteredSubtopicList(IUAElement navigationElement) {
		IUAElement[] children = navigationElement.getChildren();
		List childTopics = new ArrayList(children.length);
		for(int i=0;i<children.length;i++) {
			IUAElement child = children[i];
			if ((child instanceof ITopic)) {
				// add topic only if it will not end up being an empty
				// container
				if ((((ITopic) child).getHref() != null && ((ITopic) child)
						.getHref().length() > 0)
						|| getFilteredSubtopicList(child).size() > 0) {
					if(FilterData.filter(request.getSession(), (ITopic)child)) {
						childTopics.add(child);
						break;
					}
					
				}
			} else {
				// it is a Toc, Anchor or Link,
				// which may have children attached to it.
				childTopics.addAll(getFilteredSubtopicList(child));
			}
		}
		return childTopics;
	}
	//**********************Filter ends**************************************

	/*
	 * Generates the content to print (the merged topics).
	 */
	public boolean generateContent(Writer out) throws ServletException, IOException {
		try {
			return generateContent(getTopic(), null, out, new HashSet(), false);
		} catch (IOException e) {
			// return resource
			synchronized (lock) {
				allowedMaxConnections++;
				lock.notifyAll();
			}
			RequestDispatcher rd = context.getRequestDispatcher("/advanced/printError.jsp"); //$NON-NLS-1$
			request.setAttribute("msg", "ioException"); //$NON-NLS-1$ //$NON-NLS-2$
			rd.forward(request, response);
			return false;
		}
		catch (Exception ex) {
			HelpWebappPlugin.logWarning("Print is terminated by an exception."); //$NON-NLS-1$
		}
	
		// return resource
		synchronized (lock) {
			allowedMaxConnections++;
			lock.notifyAll();
		}
		
		return true;
	}
	
	
	/*
	 * Auxiliary method for recursively generating print content.
	 */
	private boolean generateContent(ITopic topic, String sectionId, Writer out, Set generated, boolean hasScript) throws IOException {
		if (topicsGenerated < allowedMaxTopics) {
			String href = topic.getHref();
			if (href != null && href.length() > 0) {
				topicsGenerated++;
				
				// get the topic content
				href = removeAnchor(href);
				String pathHref = href.substring(0, href.lastIndexOf('/') + 1);
			//	String baseHref = "http://" + request.getServerName() + ":" + request.getServerPort() + request.getContextPath() + "/topic" + pathHref;   //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
				String baseHref = "../topic" + pathHref;   //$NON-NLS-1$
				
				String content;
				if (!generated.contains(href)) {
					generated.add(href);
					
					String extension = (href.substring(
							href.lastIndexOf(".") + 1, href.length())).toLowerCase(Locale.ENGLISH);
					if (extension != null && extension.length() > 0) {
						String url = "http://" + request.getServerName() + ":" + request.getServerPort() + request.getContextPath() + "/topic" + href;
						if (extension.indexOf("htm") != -1) {			//.htm(l) or .xhtm(l)
							content = getContent(href, locale);	
							if(!hasScript)
								hasScript = hasJSCode(content);
							
							transferContent(content, url, baseHref, sectionId, out);
						}
						else if (images.contains(extension)) {			// images file						
							generateImageFile(url, sectionId, out);
						}
						else {											// .pdf or other unsupported file
							generateUnsupportedFile(url, sectionId, out);
						}
					}
				}							
			}
			if (!isSinglePrint()) {
				ITopic[] subtopics = EnabledTopicUtils.getEnabled(topic.getSubtopics());
				int contentID = 1;
				for (int i=0;i<subtopics.length;++i) {
					// filter the topic first
					if (!FilterData.filter(request.getSession(), subtopics[i])) {
						continue;
					}
					
					if(subtopics[i].getHref() == null || subtopics[i].getHref().length() == 0) {
						ITopic[] subsubtopics = getFilteredSubtopics(subtopics[i]);
						if(subsubtopics.length == 0) {
							continue;
						}
					}
					//filter ends
					String subsectionId = (sectionId != null ? sectionId + "." : "") + contentID++; //$NON-NLS-1$ //$NON-NLS-2$
					hasScript = generateContent(subtopics[i], subsectionId, out, generated, hasScript);
				}
			}
			
			
		}
		return hasScript;
	}
	
	public boolean transferContent(String content, String url, String baseHref, String sectionId, Writer out) throws IOException {	
		// fixed for P000851 1/2: JavaScript appears in print preview
		List chunks = RobustHTMLParser.parse(content);//
		// output string buffer
		StringBuffer output = new StringBuffer();
		int pieceNumber = chunks.size();
		Chunk chunk;
		for (int i = 0; i < pieceNumber; ++i) {
			chunk = (Chunk) chunks.get(i);
			if (chunk instanceof CommentTag) {
				// do nothing
			// uat4i00000210
			} else if ((chunk instanceof Tag) && ("script".equals(((Tag) chunk).getTag()))) { //$NON-NLS-1$) 
				// do nothing
			// end of uat4i00000210
			} else {
				if (chunk instanceof OpenTag) {
					if (!filterOpenTag((OpenTag) chunk)) {						
						generateUnsupportedFile(url, sectionId, out);
						return false;
					}
				}
				// append the chunk
				output.append(chunk.toString());
			}
		}
		
		content = output.toString();//
		if (sectionId != null) {
			content = injectHeading(content, sectionId);
		}
		content = normalizeHrefs(content, baseHref);
		out.write(content);	
		
		return true;
	}
	
	//fixed for P000851 1/2
	private static boolean filterOpenTag(OpenTag tag) {
		String tagName = tag.getTag();

		if ("meta".equals(tagName)) { //$NON-NLS-1$
			String value = (String) tag.getParams().get("http-equiv"); //$NON-NLS-1$
			if ((value != null) && "refresh".equals(value)) { //$NON-NLS-1$
				return false;
			}
		}
		else if ("frameset".equals(tagName)) { //$NON-NLS-1$
			return false;
		}

		return true;
	}
	
	public void generateImageFile(String url, String sectionId, Writer out) throws IOException {			
		out.write("<html>\n"); 
		out.write("<body>\n"); 
		out.write("<div dir=\"" + direction + "\">\n");
		out.write("<p>\n");
		if (sectionId != null && sectionId.length() > 0) {
			out.write("<a id=\"section" + sectionId + "\">" + "</a>");
		}
		out.write("<img src=\"" + url + "\">\n"); 
		out.write("</p>\n");
		out.write("</div>");
		out.write("</body>\n"); 
		out.write("</html>\n"); 
		out.flush();
	}
	
	public void generateUnsupportedFile(String url, String sectionId, Writer out) throws IOException {
		topicsGenerated--;
		
	//	out.write("<html>\n"); 
	//	out.write("<body>\n"); 
		out.write("<div dir=\"" + direction + "\">\n");
		out.write("<p>\n");
		if (sectionId != null && sectionId.length() > 0) {
			out.write("<a id=\"section" + sectionId + "\">" + "</a>");
		}
		out.write(ServletResources.getString("print_not_support", request)); 
		out.write("</p>\n");
		out.write("</div>");
	//	out.write("</body>\n"); 
	//	out.write("</html>\n"); 
		out.flush();
	}
	// Story 702: quick print enhancement ends

	/*
	 * Injects the sectionId into the document heading.
	 */
	private String injectHeading(String content, String sectionId) {
		Matcher matcher = PATTERN_HEADING.matcher(content);
		if (matcher.find()) {
			String heading = "<a id=\"section" + sectionId + "\">" + sectionId + ". </a>"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			return content.substring(0, matcher.start(1)) + heading + content.substring(matcher.start(1));
		}
		return content;
	}
	
	/*
	 * Normalizes all external links since we're not at the same base href as the
	 * topics we're printing.
	 */
	private String normalizeHrefs(String content, String baseHref) {
		StringBuffer buf = new StringBuffer();
		Matcher matcher = PATTERN_LINK.matcher(content);
		int prev = 0;
		while (matcher.find()) {
			buf.append(content.substring(prev, matcher.start(2)));
			buf.append(baseHref);
			buf.append(matcher.group(2));
			prev = matcher.end();
		}
		buf.append(content.substring(prev));
		return buf.toString();
	}
	
	/*
	 * Returns the string content of the referenced topic in UTF-8.
	 */
	private String getContent(String href, String locale) {
		InputStream in = HelpSystem.getHelpContent(href, locale);
		StringBuffer buf = new StringBuffer();
		if (in != null) {
			try {
				String charset = HTMLDocParser.getCharsetFromHTML(in);
				if (charset == null) {
					charset = "UTF-8"; //$NON-NLS-1$
				}
				InputStream rawInput = HelpSystem.getHelpContent(href, locale);
//				in = DynamicXHTMLProcessor.process(href, rawInput, locale, RemoteHelp.isAllowed());
				boolean filter = BaseHelpSystem.getMode() != BaseHelpSystem.MODE_INFOCENTER;
				in = DynamicXHTMLProcessor.process(href, rawInput, locale, filter);

				// modify for quick print for anchor doc
				if(in == null) {
					//in.close();
					in = HelpSystem.getHelpContent(href, locale);
				}
				//ends
				Reader reader = new BufferedReader(new InputStreamReader(in, charset));
				char[] cbuf = new char[4096];
				int num;
				while ((num = reader.read(cbuf)) > 0) {
					buf.append(cbuf, 0, num);
				}
			}
			catch (Exception e) {
				String msg = "Error retrieving print preview content for " + href; //$NON-NLS-1$
				HelpWebappPlugin.logError(msg, e);
			}
			finally {
				try {
					in.close();
				}
				catch (Exception e) {}
			}
		}
		return buf.toString();
	}

	/*
	 * Returns the toc containing the selected topic(s).
	 */
	private IToc getToc() {
		String tocParam = request.getParameter("toc"); //$NON-NLS-1$
		if (tocParam != null && tocParam.length() > 0) {
			return HelpPlugin.getTocManager().getToc(tocParam, getLocale());
		}
		String topicParam = request.getParameter("topic"); //$NON-NLS-1$
		if (topicParam != null && topicParam.length() > 0) {
			if (topicParam.startsWith("/../nav/")) { //$NON-NLS-1$
				String navPath = topicParam.substring(8);
				StringTokenizer tok = new StringTokenizer(navPath, "_"); //$NON-NLS-1$
				int index = Integer.parseInt(tok.nextToken());
				return HelpPlugin.getTocManager().getTocs(getLocale())[index];
			}
			IToc[] tocs = HelpPlugin.getTocManager().getTocs(getLocale());
			for (int i=0;i<tocs.length;++i) {
				if (tocs[i].getTopic(topicParam) != null) {
					return tocs[i];
				}
			}
		}
		return null;
	}

	/*
	 * Returns the selected topic.
	 */
	private ITopic getTopic() {
		String topicParam = request.getParameter("topic"); //$NON-NLS-1$
		String anchorParam = request.getParameter("anchor"); //$NON-NLS-1$
		if (anchorParam!=null) {
			topicParam = topicParam + '#' + anchorParam;
		}
		if (topicParam != null && topicParam.length() > 0) {
			if (topicParam.startsWith("/../nav/")) { //$NON-NLS-1$
				String navPath = topicParam.substring(8);
				StringTokenizer tok = new StringTokenizer(navPath, "_"); //$NON-NLS-1$
				int index = Integer.parseInt(tok.nextToken());
				ITopic topic = HelpPlugin.getTocManager().getTocs(getLocale())[index].getTopic(null);
				while (tok.hasMoreTokens()) {
					index = Integer.parseInt(tok.nextToken());
					topic = topic.getSubtopics()[index];
				}
				return topic;
			}
			else {
				IToc[] tocs = HelpPlugin.getTocManager().getTocs(getLocale());
				for (int i=0;i<tocs.length;++i) {
					ITopic topic = tocs[i].getTopic(topicParam);
					if (topic != null) {
						return topic;
					}
					// Test for root node as topic
					topic = tocs[i].getTopic(null);
					if (topicParam.equals(topic.getHref())) {
						return topic;
					}
				}
			}
			return null;
		}
		return getToc().getTopic(null);
	}
	
	private static String removeAnchor(String href) {
		int index = href.indexOf('#');
		if (index != -1) {
			return href.substring(0, index);
		}
		return href;
	}
	
	public boolean hasJSCode(String content) {
		if(content.indexOf("<script")>0)
			return true;
		
		return false;
	}
}
